/*Generacja sinusa*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define Fs    8000
#define Nbits  16      // liczba bitow probek (np. Nbits=16, probki z zakresu <-2^15,2^15>)
#define FILTER_LENGHT 105

#define GENERATOR_PRINT  1
#define SUMATOR_PRINT    2
#define GEORTZEL_PRINT   4

#define SIN1_PLOT        8
#define SIN2_PLOT        16
#define SUM_PLOT         32
#define S_N_PLOT         64
#define POWER_PLOT       128

#define MSG_PRINT POWER_PLOT+S_N_PLOT+SUM_PLOT+SIN2_PLOT+SIN1_PLOT+SUMATOR_PRINT+GEORTZEL_PRINT+GENERATOR_PRINT 

#define GNUPLOT  //zakomentowac jak nie ma w sytemie GNUPLOTA

#ifdef GNUPLOT
#include "gnuplot_i.h"  // wrysowanie wykresow
#endif


const float 	PI = 3.14159265358979323846f;
void plot_samples(short int*  waveform_in, int liczba_probek, char *tutul);
void gen_sin(int liczba_bitow, int fsampling, int fout, short int*  sin_out, int liczba_probek);
void sum_sin(short int*  sin1_in,short  int*  sin2_in,short  int * sum_sin_out, int liczba_probek);
short int geortzl(int liczba_bitow,int N, int k,short  int* signal_in, short int* signal_out);

int main()
{
    
    int i;
    
    short int dtmf_freq_tab[8]={697, 770, 852, 941, 1209, 1336, 1477, 1633};
    
    short int k;
   
    short int tab_sin1[FILTER_LENGHT]; 
    short int tab_sin2[FILTER_LENGHT];
    short int tab_sum_sin[FILTER_LENGHT];    
    short int power[FILTER_LENGHT]={0};
    short int tab_s_out[FILTER_LENGHT];
    char plot_title[64]={0};
    
    gen_sin(Nbits-2, Fs, dtmf_freq_tab[1], tab_sin1, FILTER_LENGHT);

#if (MSG_PRINT & SIN1_PLOT)
    plot_samples(tab_sin1, FILTER_LENGHT, "sin f1");    
#endif
    
    gen_sin(Nbits-2, Fs, dtmf_freq_tab[3], tab_sin2, FILTER_LENGHT); 

#if (MSG_PRINT & SIN2_PLOT)
    plot_samples(tab_sin2, FILTER_LENGHT, "sin f2");    
#endif
    
    sum_sin(tab_sin1, tab_sin2, tab_sum_sin, FILTER_LENGHT);

#if (MSG_PRINT & SUM_PLOT)
    plot_samples(tab_sum_sin, FILTER_LENGHT, "suma f1+f2");    
#endif
    
    for(i=0; i<8; i++)
    {
        k= round(1.0*FILTER_LENGHT*dtmf_freq_tab[i]/Fs);
        power[i]=geortzl(Nbits, FILTER_LENGHT, k, tab_sum_sin, tab_s_out);
#if (MSG_PRINT & S_N_PLOT)
        sprintf(plot_title, "detect DTMF %i Hz (k=%i)       s(n)=x(n)+A*s(n-1)-s(n-2)", dtmf_freq_tab[i], k);
        plot_samples(tab_s_out, FILTER_LENGHT, plot_title); 
#endif
    }
#if (MSG_PRINT & POWER_PLOT)
    plot_samples(power, 9, "DTMF Power       P=s(N)^2+s(N-1)^2-A*s(N)*s(N-1)"); 
#endif
    return 0;                   
}

void gen_sin(int liczba_bitow, int fsampling, int fout, short int*  sin_out, int liczba_probek)
{
    int i;
    short int y[3]={0};
    int tmp_mul_result;
    double N=(double)(1.0*fsampling/fout);
    int skala=pow(2,liczba_bitow);              
    short int A=round(2.0*cos( 2 * PI/N)*skala); // UWAGA A z zakresu <-2:2>
    short int C=round(1.0*sin( 2 * PI/N)*skala); 
    y[1]=C;

#if (MSG_PRINT & GENERATOR_PRINT)
    printf("***************************************************************************\n");
    printf("***\tGenerator przebiegu sinusoidalnego o parametrach:            \t***\n");
    printf("***\tfout=%i Hz, A=%i, C=%i, N=%f, skala=%i        \t***\n", fout, A, C, N, skala);
    printf("***\ty(n)=(A*y(n-1))/skala-y(n-2)=(%i*y(n-1))/%i-y(n-2)    \t***\n", A, skala);
    printf("***************************************************************************\n");
    printf("---------------------------------------------------------------\n");
    printf("|n       y(n-1)      A*y(n-1)  A*y(n-1)/skala    y(n-2)   y(n)|\n");
    printf("---------------------------------------------------------------\n");
#endif
    
    for(i=0; i<liczba_probek; i++)
    {
        y[2]=y[1];    //aktualizacja wartosci probki n-2  
	y[1]=y[0];     //aktualizacja wartosci probki n-1
        
        
        // wynik aktualnej probki wylicza sie z nastepujacego rownania y(k) = ((A*skala)* y(k-1))/skala - y(k-2)
        // rownanie to zostalo rozpisane tak aby w jednej linii wykonywana byla jedna operacja
        //(mam nadziej ze zwiekszy to czytelnosc tego zapisu)
        tmp_mul_result= A * y[1];  // nalezy pomnozyc probke n-1 i przeskalowany wsp A - UWAGA wynik 32 bitowy

#if (MSG_PRINT & GENERATOR_PRINT)
        printf("|%i\t%6i\t%11i\t", i, y[1], tmp_mul_result);
#endif
        
        y[0]= round(1.0*tmp_mul_result/skala);//wynik mnozenie nalezy przeskalowac (podzielic przez skale)
        
#if (MSG_PRINT & GENERATOR_PRINT)
        printf("%6i\t\t%6i\t", y[0],y[2]);
#endif

        y[0]= y[0] - y[2];             // od przeskalowanego wyniku mnozenia odjac probke n-2

#if (MSG_PRINT & GENERATOR_PRINT)
        printf("%6i|\n", y[0]);
#endif
        *sin_out=y[0];
        sin_out++;
        
    }
#if (MSG_PRINT & GENERATOR_PRINT)
    printf("---------------------------------------------------------------\n\n");
#endif

}
short int geortzl(int liczba_bitow,int N, int k, short int* signal_in,short  int* signal_out)
{
    int i;
    int skala=(int)pow(2,liczba_bitow-4);
    int s[3]={0};  //UWAGA wynik wyjscowe z IIR'a moga przekraczac zakres wejsciowy
                             //(np. dla zakresu 16bitowego i N=105 wyjscie filtru moze byc 22bitowe)
                             // najbardziej prawdopodone miejsce wystapienia przepelnienia to
                             // operacja mnozenienia y(n-1) i stała A ( A moze byc np 12bitowe)
    short int A=(short int)round(2.0*cos( 2 * PI*k/N)*skala); 
    short int power=0;
    float power_f=0.0;
    
#if (MSG_PRINT & GEORTZEL_PRINT)
    printf("***************************************************************************\n");
    printf("***\tDetektor przebiegu sinusoidalnego o parametrach:                ***\n");
    printf("***\tk=%i,N=%i, A=%i, skala=%i                               \t***\n",k,N, A, skala);
    printf("***\ts(n)=x(n)+(A*s(n-1))/skala-s(n-2)                       \t***\n");
    printf("***************************************************************************\n");
    printf("--------------------------------------------------------------------------------\n");
    printf("|n       s(n-1)     A*s(n-1)     A*s(n-1)/s  s(n-2)         x(n)          s(n) |\n", A);
    printf("---------------------------------------------------------------------------------\n");
#endif

    for(i=0; i<N; i++)
    {
        s[2]=s[1];    //aktualizacja wartosci probki n-2  
	s[1]=s[0];     //aktualizacja wartosci probki n-1
        
        
 // wynik aktualnej probki wylicza sie z nastepujacego rownania s(k) = x(k)+((A* s(k-1))/skala - s(k-2)
        s[0]= A*s[1];              // nalezy pomnozyc probke n-1 i przeskalowany wsp A

#if (MSG_PRINT & GEORTZEL_PRINT)        
        printf("|%i\t%6i\t%11i\t", i, s[1], s[0]);
#endif

        s[0]= round(1.0*s[0]/skala);  //wynik mnozenie nalezy przeskalowac w dol(podzielic przez skale)

#if (MSG_PRINT & GEORTZEL_PRINT)
        printf("%6i\t%9i\t%9i\t", s[0], s[2], *signal_in);
#endif

        // od przeskalowanego wyniku mnozenia odjac probke n-2 i dodac probke wej
        // mozna rowniez przeskalowac probki wejsciowe co ograniczy liczbe bitów wyniku
        s[0]= round(1.0*(*signal_in)/pow(2,4)) + s[0] -s[2];

#if (MSG_PRINT & GEORTZEL_PRINT)
        printf("%6i |\n",s[0]);
#endif

        *signal_out=s[0]/pow(2,8);
        signal_out++;
        signal_in++;
    }
    //zabezpieczenie przed przekroczeniem zakresu(np. 16bitowego -> (2^16)^2 + (2^16)^2 = 2^33)
    // aby nie przekroczys zakresu 16bitowego mozna mnozyc liczby max 7-bitowe 
    s[0]=round(1.0*s[0]/pow(2,13)); 
    s[1]=round(1.0*s[1]/pow(2,13));
    

    // moc sygnalu
    power=s[0]*s[0] + s[1]*s[1] - round(1.0*(((A*s[0])/skala)*s[1]));   

#if (MSG_PRINT & GEORTZEL_PRINT)
    printf("|------------------------------------------------------------------------------|\n");
    printf("|\tpower=s(N)^2 + s(N-1)^2- A*s(N)*s(N-1)=%i                               |\n",power);
    printf("|------------------------------------------------------------------------------|\n");
                  
#endif

    return power;
}

void sum_sin(short int*  sin1_in, short int*  sin2_in,short int * sum_sin_out, int liczba_probek)
{
    int i;
#if (MSG_PRINT & SUMATOR_PRINT)
    printf("*******************************************\n");
    printf("***\tSuma dwoch sygnalow okresowych  ***\n");
    printf("*******************************************\n");
    printf("-----------------------------------------\n");
    printf("|n        y1(n)  y2(n) y(n)=y1(n)+y2(n)  |\n");
    printf("-----------------------------------------\n");
#endif
    
    for(i=0; i<liczba_probek; i++)
    {
        *sum_sin_out=*sin1_in + *sin2_in;
#if (MSG_PRINT & SUMATOR_PRINT)
        printf("|%i\t%6i\t%6i\t%12i     |\n", i, *sin1_in, *sin2_in, *sum_sin_out);
#endif
        sum_sin_out++;
        sin1_in++;
        sin2_in++;
        
    }
#if (MSG_PRINT & SUMATOR_PRINT)
        printf("|----------------------------------------|\n", i, *sum_sin_out, *sum_sin_out);
#endif

}

void plot_samples(short int*  waveform_in, int liczba_probek , char *tytul)
{
    #ifdef GNUPLOT
    int i;
    double *tab_plot;

    gnuplot_ctrl *h1 ;
    h1 = gnuplot_init();

    tab_plot= (double*) malloc(liczba_probek * sizeof *tab_plot);

    
    //printf("Rysowanie przebiegu sinusoidalnego\n");
    for(i=0; i<liczba_probek; i++)
    {
        tab_plot[i]=(double)*waveform_in;
        waveform_in++;
    }
    // wykres w gnuplot
    //gnuplot_setstyle(h1, "lines") ;
    gnuplot_setstyle(h1, "steps") ;
    //gnuplot_setstyle(h1, "impulses") ;
    gnuplot_plot_x(h1, tab_plot ,liczba_probek+1, tytul); //pojedynczy sygnal

    //gnuplot_hardcopy_colour(h1, "sine.png"); // zapis do pliku graficznego 
 
    gnuplot_cmd(h1, "pause mouse any"); // czeka na nacisniecie dowolnego klawisza

    sleep(1);
    gnuplot_close(h1);
    
    free(tab_plot);
    #endif
}
